local mod = LibStub("AceAddon-3.0"):NewAddon("InternalCooldowns", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0")

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")
local callbacks = CallbackHandler:New(mod)

local procDB, cooldowns

local GetInventoryItemLink = _G.GetInventoryItemLink
local substr = _G.string.sub
local wipe = _G.wipe

local slots = {
	[0] = "CharacterAmmoSlot",
	[1] = "CharacterHeadSlot",
	[2] = "CharacterNeckSlot",
	[3] = "CharacterShoulderSlot",
	[4] = "CharacterShirtSlot",
	[5] = "CharacterChestSlot",
	[6] = "CharacterWaistSlot",
	[7] = "CharacterLegsSlot",
	[8] = "CharacterFeetSlot",
	[9] = "CharacterWristSlot",
	[10]= "CharacterHandsSlot",
	[11]= "CharacterFinger0Slot",
	[12]= "CharacterFinger1Slot",
	[13]= "CharacterTrinket0Slot",
	[14]= "CharacterTrinket1Slot",
	[15]= "CharacterBackSlot",
	[16]= "CharacterMainHandSlot",
	[17]= "CharacterSecondaryHandSlot",
	[18]= "CharacterRangedSlot",
	[19]= "CharacterTabardSlot"
}

local spellToItem = {}
local equippedItems = {}
local playerGUID

function mod:OnInitialize()
	procDB, cooldowns = self.procDB, self.cooldowns
	
	for k, v in pairs(procDB) do
		local e = spellToItem[v]
		if e then
			if type(e) == "table" then
				tinsert(k, e)
			else
				spellToItem[v] = {e, k}
			end
		else
			spellToItem[v] = k
		end
	end
end

function mod:OnEnable()
	playerGUID = UnitGUID("player")
	
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
	self:RawHook("GetInventoryItemCooldown", true)
	self:RawHook("GetActionCooldown", true)
	
	self:UNIT_INVENTORY_CHANGED("player")
end

function mod:OnDisable()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("UNIT_INVENTORY_CHANGED")
	self:Unhook("GetInventoryItemCooldown")
	self:Unhook("GetActionCooldown", true)
end

local INVALID_EVENTS = {
	SPELL_DISPEL 			= true,
	SPELL_DISPEL_FAILED 	= true,
	SPELL_STOLEN 			= true,
	SPELL_AURA_REMOVED 		= true,
	SPELL_AURA_REMOVED_DOSE = true,
	SPELL_AURA_BROKEN 		= true,
	SPELL_AURA_BROKEN_SPELL = true,
	SPELL_CAST_FAILED 		= true
}

function mod:COMBAT_LOG_EVENT_UNFILTERED(frame, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellID, spellName)
	if ((destGUID == playerGUID and (sourceGUID == nil or sourceGUID == destGUID)) or sourceGUID == playerGUID) and not INVALID_EVENTS[event] and substr(event, 0, 6) == "SPELL_" then
		local itemID = spellToItem[spellID]
		if itemID then
			if type(itemID) == "table" then
				for k, v in ipairs(itemID) do
					self:SetCooldownFor(v, spellID)
				end
			else
				self:SetCooldownFor(itemID, spellID)
			end
		end
	end
end

function mod:UNIT_INVENTORY_CHANGED(unit)
	if unit == "player" then
		wipe(equippedItems)
		for i = 1, 19 do
			local link = GetInventoryItemLink(unit, i)
			if link then
				equippedItems[tonumber(link:match("item:(%d+)") or 0)] = i
			end			
		end
	end
end

local cooldownStartTimes, cooldownDurations = {}, {}

function mod:SetCooldownFor(itemID, spellID)
	local slot = equippedItems[itemID]
	if slot then
		local duration = cooldowns[itemID] or 45
		cooldownStartTimes[slot] = GetTime()
		cooldownDurations[slot] = duration
		local start = GetTime()
		_G[slots[slot] .. "Cooldown"]:SetCooldown(start, duration)
		
		callbacks:Fire("InternalCooldowns_Proc", itemID, spellID, start, duration)
	end
end

function mod:GetInventoryItemCooldown(unit, slot)
	local start, duration, enable = self.hooks.GetInventoryItemCooldown(unit,slot)
	if not enable or enable == 0 then
		if cooldownStartTimes[slot] and cooldownDurations[slot] then
			return cooldownStartTimes[slot], cooldownDurations[slot], 1
		end
	end
	return start, duration, enable
end

function mod:GetActionCooldown(slotID)
	local t, id, subtype, globalID = GetActionInfo(slotID)
	if t == "item" then
		local slot = equippedItems[id]
		if cooldownStartTimes[slot] and cooldownDurations[slot] then
			return cooldownStartTimes[slot], cooldownDurations[slot], 1
		end
	end
	return self.hooks.GetActionCooldown(slotID)
end
